package com.qhtx.eta.domain.job;

import com.alibaba.nacos.api.utils.StringUtils;
import com.google.common.collect.Lists;
import com.qhtx.eta.common.eunms.DataSourceType;
import com.qhtx.eta.common.eunms.ErrorEnum;
import com.qhtx.eta.common.exception.ETAException;
import com.qhtx.eta.domain.annotation.DomainTransDataSource;
import com.qhtx.eta.domain.config.ScheduleConfig;
import com.qhtx.eta.domain.config.ScheduleTaskConfig;
import com.qhtx.eta.domain.convert.ClassifyMappingDTOConverter;
import com.qhtx.eta.domain.entity.*;
import com.qhtx.eta.domain.enums.ETADataStatus;
import com.qhtx.eta.domain.enums.ETAInterfaceEnum;
import com.qhtx.eta.domain.enums.ETATaskEnum;
import com.qhtx.eta.domain.enums.ExcludedQuotaSource;
import com.qhtx.eta.domain.facade.ETAFacadeService;
import com.qhtx.eta.domain.service.DWIndexFrameService;
import com.qhtx.eta.domain.service.ETAClassifyService;
import com.qhtx.eta.domain.service.ETAPushDataService;
import com.qhtx.eta.domain.service.ETAQuotaService;
import com.qhtx.eta.domain.api.ApiServiceHolder;
import com.qhtx.eta.domain.api.http.request.ETADataHttpRequest;
import com.qhtx.eta.common.constant.ETAConstants;
import com.qhtx.eta.domain.task.FixedDelayTaskRegistrar;
import com.qhtx.eta.domain.task.SchedulingRunnable;
import com.qhtx.eta.infra.annotation.UseDataSource;
import com.qhtx.eta.infra.entity.dw.TEtaIndexPushTask;
import com.qhtx.eta.infra.enums.Frequency;
import lombok.extern.slf4j.Slf4j;
import org.redisson.api.RLock;
import org.redisson.api.RedissonClient;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.boot.CommandLineRunner;
import org.springframework.stereotype.Component;
import org.springframework.util.CollectionUtils;

import javax.annotation.Resource;
import java.text.SimpleDateFormat;
import java.time.LocalDate;
import java.time.format.DateTimeFormatter;
import java.util.*;
import java.util.concurrent.*;
import java.util.stream.Collectors;


@Component
@Slf4j
public class ETASyncJob implements CommandLineRunner {

    @Value("${eta.http.interval}")
    private int interval;

    @Value("${eta.http.numberOfPeriods}")
    private int numberOfPeriods;
    @Resource
    private ApiServiceHolder apiServiceHolder;

    @Resource
    private ETAQuotaService etaQuotaService;
    @Resource
    private ETAClassifyService etaClassifyService;

    @Resource(name = "ETADataThreadPool")
    private ThreadPoolExecutor etaDataThreadPool;
    @Resource(name = "IndexFrameThreadPool")
    private ThreadPoolExecutor indexFrameThreadPool;
    @Resource
    private FixedDelayTaskRegistrar fixedDelayTaskRegistrar;

    @Resource
    private ScheduleTaskConfig scheduleTaskConfig;


    @Resource
    private DWIndexFrameService dwIndexFrameService;

    @Resource
    private ETAFacadeService facadeService;
    @Resource(name = "IndexDataThreadPool")
    private ThreadPoolExecutor indexDataThreadPool;
    @Resource
    private ETAPushDataService etaPushDataService;

    @Resource
    private RedissonClient redissonClient;

    private static final Map<Integer, ETATaskEnum> taskMappingMap = new ConcurrentHashMap<>(16);

    public ETATaskEnum getTaskType(int configId) {
        return taskMappingMap.get(configId);
    }

    /**
     * 同步指标数据到自有数据源
     */
    public void syncETAPushData() {
        //获取需要更新的数据
        List<DWIndexDTO> pushTaskList = etaPushDataService.getPushIndexTaskList();
        if (!CollectionUtils.isEmpty(pushTaskList)) {
            log.info("更新任务列表:{}", pushTaskList.size());
            pushTaskList.forEach(item -> {
                boolean isLock;
                RLock rLock = redissonClient.getLock(item.getIndexCode());
                isLock = rLock.tryLock();
                if (isLock) {
                    try {
                        DWIndexDTO dwIndexDTO = dwIndexFrameService.getIndex(item.getIndexCode());
                        //如果数据没有直接初始化
                        if (StringUtils.isBlank(item.getLatestDataDate())) {
                            log.info("初始化指标:{}", item.getIndexCode());
                            facadeService.pushIndex(item.getIndexCode(), false);
                            return;
                        }
                        //提交给线程池去完成，每个指标顺序执行不并发。
                        CompletableFuture<Void> future = CompletableFuture.runAsync(() -> {
                            //获取指标数据
                            List<DWIndexDataDTO> dataList = etaPushDataService.getUpdateDataList(item.getIndexCode(), item.getLatestDataDate());
                            dataList.sort(Comparator.comparingLong(o -> o.getDate().getTime()));
                            if (!CollectionUtils.isEmpty(dataList)) {
                                dwIndexDTO.setDataList(dataList);
                                //api数据发送
                                if (dataList.size() > ETAConstants.ETA_PUSH_DATA_LIMIT) {
                                    log.info("数据量超过限制，分批更新推送");
                                    Lists.partition(dataList, ETAConstants.ETA_PUSH_DATA_CHUNK).forEach(list -> {
                                        Date latestDataDate = list.stream().map(DWIndexDataDTO::getDate).max(Comparator.comparing(Date::getTime)).get();
                                        DWIndexDTO dataDTO = new DWIndexDTO();
                                        dataDTO.setIndexCode(dwIndexDTO.getIndexCode());
                                        dataDTO.setFrequency(dwIndexDTO.getFrequency());
                                        dataDTO.setIndexName(dwIndexDTO.getIndexName());
                                        dataDTO.setRemark(dwIndexDTO.getRemark());
                                        dataDTO.setSourceName(dwIndexDTO.getSourceName());
                                        dataDTO.setUnit(dwIndexDTO.getUnit());
                                        dataDTO.setDataList(list);
                                        List<?> result = apiServiceHolder.runApi(ETAInterfaceEnum.PUSH_ETA_INDEX, dataDTO);
                                        if (result == null) {
                                            log.error("调用ETA接口失败，停止更新");
                                            throw new ETAException(ErrorEnum.PUSH_TO_ETA_ERROR);
                                        }
                                        try {
                                            Thread.sleep(1000);
                                        } catch (InterruptedException e) {
                                            log.warn("线程中断异常");
                                        }
                                        TEtaIndexPushTask tEtaIndexPushTask = etaPushDataService.getETAPushTaskByIndexCode(dataDTO.getIndexCode());
                                        if (tEtaIndexPushTask == null) {
                                            log.error("未找到对应指标任务记录:{}", dataDTO.getIndexCode());
                                            throw new ETAException(ErrorEnum.INDEX_UPDATE_ERROR);
                                        }
                                        tEtaIndexPushTask.setLatestDataDate(new SimpleDateFormat(ETAConstants.DATE_PATTERN).format(latestDataDate));
                                        int currentCount = tEtaIndexPushTask.getDataCount() == null ? 0 : tEtaIndexPushTask.getDataCount();
                                        tEtaIndexPushTask.setDataCount(currentCount + list.size());
                                        tEtaIndexPushTask.setUpdateTime(new Date());
                                        tEtaIndexPushTask.setLastPushTime(new Date());
                                        etaPushDataService.updateTaskByIndexCode(tEtaIndexPushTask);
                                    });
                                } else {
                                    apiServiceHolder.runApi(ETAInterfaceEnum.PUSH_ETA_INDEX, dwIndexDTO);
                                    TEtaIndexPushTask tEtaIndexPushTask = etaPushDataService.getETAPushTaskByIndexCode(dwIndexDTO.getIndexCode());
                                    if (tEtaIndexPushTask == null) {
                                        log.error("未找到对应指标任务记录:{}", dwIndexDTO.getIndexCode());
                                        throw new ETAException(ErrorEnum.INDEX_UPDATE_ERROR);
                                    }
                                    Date latestDataDate = dwIndexDTO.getDataList().stream().map(DWIndexDataDTO::getDate).max(Comparator.comparing(Date::getTime)).get();
                                    tEtaIndexPushTask.setLatestDataDate(new SimpleDateFormat(ETAConstants.DATE_PATTERN).format(latestDataDate));
                                    int currentCount = tEtaIndexPushTask.getDataCount() == null ? 0 : tEtaIndexPushTask.getDataCount();
                                    tEtaIndexPushTask.setDataCount(currentCount + dwIndexDTO.getDataList().size());
                                    tEtaIndexPushTask.setUpdateTime(new Date());
                                    tEtaIndexPushTask.setLastPushTime(new Date());
                                    etaPushDataService.updateTaskByIndexCode(tEtaIndexPushTask);
                                }
                            }
                        }).thenRun(() -> {
                            log.info("更新指标任务表完成{}", item.getIndexCode());
                        });
                        future.exceptionally(exception -> {
                            log.error("同步指标任务表失败:", exception);
                            return null;
                        });
                        future.get();
                    } catch (Exception e) {
                        log.info("更新数据失败", e);
                    } finally {
                        rLock.unlock();
                        log.info("释放锁{}", item.getIndexCode());
                    }
                } else {
                    log.info("未获取到锁，跳过更新{}", item.getIndexCode());
                }
            });
        }
    }

    /**
     * 同步指标信息
     */
    public void syncETAInfo() {
        List<?> dataList = apiServiceHolder.runApi(ETAInterfaceEnum.GET_EDB_LIST);
        if (dataList != null) {
            List<QuotaInfoDTO> etaQuotaList = dataList.stream().filter(item -> item instanceof QuotaInfoDTO && Arrays.stream(ExcludedQuotaSource.values()).noneMatch(excludedSource -> ((QuotaInfoDTO) item).getEdbSource().equals(excludedSource.getSourceId()))).map(item -> (QuotaInfoDTO) item).collect(Collectors.toList());
            log.info("================================ 开始同步接口指标列表 ================================ ");
            log.info("================================ 指标接口列表数量:{}条================================ ", etaQuotaList.size());
            etaQuotaService.syncQuota(etaQuotaList);
            log.info("================================ 同步接口指标列表完成 ================================ ");
        } else {
            log.warn("调用接口失败,无数据同步");
        }
    }

    /**
     * 同步ETA的指标框架到一期
     */
    public void syncETAClassifyList() {
        List<?> respList = apiServiceHolder.runApi(ETAInterfaceEnum.GET_CLASSIFY_LIST);
        if (respList != null) {
            List<ClassifyMappingDTO> classifyList = respList.stream().filter(item -> item instanceof ClassifyMappingDTO).map(item -> (ClassifyMappingDTO) item).collect(Collectors.toList());
            log.info("================================ 开始同步接口指标分类列表 ================================ ");
            log.info("================================ 接口指标分类列表数量:{}条================================ ", classifyList.size());
            try {
                etaClassifyService.syncClassify(classifyList);
                // 获取信号量
                List<ClassifyMappingDTO> processingList = etaClassifyService.getProcessingList();
                if (!CollectionUtils.isEmpty(processingList)) {
                    log.info("processingList size:{}", processingList.size());
                    dwIndexFrameService.processing(ClassifyMappingDTOConverter.INSTANCE.convertDTOToDWIndexFrameList(processingList));
                }
                //开始获取新增和需要更新的框架数据
            } catch (Exception e) {
                e.printStackTrace();
                log.warn("同步eta指标分类失败:{}", e.getMessage());
            }
            log.info("================================ 同步接口指标分类列表完成 ================================ ");
        } else {
            log.warn("调用接口结束,无数据同步");
        }
    }


    /**
     * 同步指标数据
     */
    public void syncETAData() {
        log.info("{} 开始同步指标数据任务 {}", ETAConstants.LINE_SEPARATOR, ETAConstants.LINE_SEPARATOR);
        Date minDate = null;
        boolean dataUpdated = false;
        QuotaInfoDTO quotaInfoDTO = new QuotaInfoDTO();
        //查询条件暂时不设置
//        quotaInfoDTO.setNoUpdate(0);
        List<QuotaDataUpdateBO> updateList = etaQuotaService.getQuotaInfoListForUpdate(quotaInfoDTO).stream().filter(item -> item.getStatus() != ETADataStatus.LATEST).collect(Collectors.toList());
        if (CollectionUtils.isEmpty(updateList)) {
            log.info("{} 当前无同步指标 {}", ETAConstants.LINE_SEPARATOR, ETAConstants.LINE_SEPARATOR);
            return;
        }
        int count = 0;
        for (QuotaDataUpdateBO quotaDataUpdateBO : updateList) {
            ETADataHttpRequest etaDataRequest = new ETADataHttpRequest();
            etaDataRequest.setUniqueCode(quotaDataUpdateBO.getUniqueCode());
            if (quotaDataUpdateBO.getStatus() == ETADataStatus.UPDATE) {
                String startDate = calculatePreviousDate(quotaDataUpdateBO.getCurrentDate(), Frequency.getFrequencyByDesc(quotaDataUpdateBO.getFrequency()), numberOfPeriods);
                etaDataRequest.setStartDate(startDate);
            }
            log.info("================================ 开始同步接口指标明细数据 ================================ ");
            List<?> dataList = apiServiceHolder.runApi(ETAInterfaceEnum.GET_EDB_DATA, etaDataRequest);
            if (CollectionUtils.isEmpty(dataList)) {
                log.info("{} 没有获取到新的指标数据 {}", ETAConstants.LINE_SEPARATOR, ETAConstants.LINE_SEPARATOR);
                continue;
            }
            List<QuotaDataDTO> quotaDataList = dataList.stream().filter(item -> item instanceof QuotaDataDTO).map(item -> {
                QuotaDataDTO data = (QuotaDataDTO) item;
                data.setUniqueCode(etaDataRequest.getUniqueCode());
                return data;
            }).collect(Collectors.toList());
//                log.info("================================ 开始同步接口指标明细数据 ================================ ");
//                if (CollectionUtils.isEmpty(quotaDataList)) {
//                    log.info("{} 没有获取到新的指标数据 {}", ETAConstants.LINE_SEPARATOR, ETAConstants.LINE_SEPARATOR);
//                    continue;
//                }
            log.info("================================ 指标接口数据[{}]数量:{}条================================ ", quotaDataList.get(0).getUniqueCode(), quotaDataList.size());
            //对指标数据做一个时间排序
            quotaDataList.sort(Comparator.comparing(QuotaDataDTO::getDataTime));
            //取最小的日期
            Date startDate = quotaDataList.get(0).getDataTime();
            minDate = getMinDate(minDate, startDate);
            count += quotaDataList.size();
            Runnable task = () -> etaQuotaService.syncData(quotaDataList);
            etaDataThreadPool.submit(task);
            pause();
        }
        log.info("{} 同步指标数据任务完成 {}", ETAConstants.LINE_SEPARATOR, ETAConstants.LINE_SEPARATOR);
        if (count > 0) {
            long minUpdateTime = 0;
            // 设置为前一天
            if (minDate != null) {
//            Calendar calendar = Calendar.getInstance();
//            calendar.setTime(minDate);
//            calendar.add(Calendar.DAY_OF_MONTH, -1);
                minUpdateTime = minDate.getTime();
            }
            etaQuotaService.LinkedCodeMessage(minUpdateTime);
        }
    }

    private Date getMinDate(Date date1, Date date2) {
        if (date2 == null) {
            return date1;
        }
        if (date1 == null) {
            return date2;
        }
        return date1.before(date2) ? date1 : date2;
    }

    private String calculatePreviousDate(String startDate, Frequency frequency, int numberOfPeriods) {
        if (frequency == null) {
            frequency = Frequency.DAILY;
        }
        LocalDate localDate = LocalDate.parse(startDate);
        DateTimeFormatter formatter = DateTimeFormatter.ofPattern(ETAConstants.DATE_PATTERN);
        switch (frequency) {
            case DAILY:
                return localDate.minusDays(numberOfPeriods).format(formatter);
            case WEEKLY:
                return localDate.minusWeeks(numberOfPeriods).format(formatter);
            case MONTHLY:
                return localDate.minusMonths(numberOfPeriods).format(formatter);
            case YEARLY:
                return localDate.minusYears(numberOfPeriods).format(formatter);
            case SEMI_ANNUAL:
                return localDate.minusMonths(6L * numberOfPeriods).format(formatter);
            case QUARTERLY:
                return localDate.minusMonths(3L * numberOfPeriods).format(formatter);
            case DECADELY:
                return localDate.minusDays(10L * numberOfPeriods).format(formatter);
            default:
                throw new IllegalArgumentException("Unsupported frequency: " + frequency);
        }
    }

    private void pause() {
        try {
            Thread.sleep(interval);
        } catch (InterruptedException e) {
            Thread.currentThread().interrupt();
            log.info("网络请求延时异常中断:[{}]", e.getLocalizedMessage());
        }
    }

    @Override
    public void run(String... args) {
        List<ScheduleConfig> list = scheduleTaskConfig.getList();
        if (!CollectionUtils.isEmpty(list)) {
            for (int index = 0; index < list.size(); index++) {
                ScheduleConfig config = list.get(index);
                ETATaskEnum etaTaskEnum = ETATaskEnum.ofName(config.getName());
                if (etaTaskEnum != null) {
                    taskMappingMap.put(index, etaTaskEnum);
                    if (config.getEnabled()) {
                        //同步指标数据任务
                        fixedDelayTaskRegistrar.addFixedDelayTask(new SchedulingRunnable(this.getClass().getName(), etaTaskEnum.getMethod(), this), config.getInterval(), config.getInitialDelay());
                    }
                }
            }
        }
    }

}
